/*
 * avrio.hpp
 * bitTest
 *
 */

#ifndef AVRIO_HPP_
#define AVRIO_HPP_

#ifndef __cplusplus
#error C++ compilation required
#endif

#include <inttypes.h>
#include <avr/io.h>
#include <avr/interrupt.h>

namespace AVRIO
{

/**
 *         
 **/
enum __bits {__bit_0=0, __bit_1=1, __bit_2=2, __bit_3=3, __bit_4=4, __bit_5=5, __bit_6=6, __bit_7=7};
enum __ports{__port_A,__port_B,__port_C,__port_D,__port_E,__port_F,__port_G};
enum __ports_reg {__reg_PORT, __reg_DDR, __reg_PIN};

//  ,      /
template <__ports port, __ports_reg reg> volatile uint8_t& __ports_ref();

//       :
// template <> inline volatile uint8_t& port_ref<port_X, reg_PORT>() { return PORTX; }
// template <> inline volatile uint8_t& port_ref<port_X, reg_DDR >() { return DDRX ; }
// template <> inline volatile uint8_t& port_ref<port_X, reg_PIN >() { return PINX ; }
//  X      ,    :
// PORTX, DDRX  PINX (,   A,  PORTA, DDRA  PINA)

#define AVRLIB_SPECIALIZE_PORT_REF(_port_) \
template <> inline volatile uint8_t& __ports_ref<__port_##_port_, __reg_PORT>() { return PORT##_port_; } \
template <> inline volatile uint8_t& __ports_ref<__port_##_port_, __reg_DDR >() { return DDR##_port_ ; } \
template <> inline volatile uint8_t& __ports_ref<__port_##_port_, __reg_PIN >() { return PIN##_port_ ; }

#if defined (PORTA)
AVRLIB_SPECIALIZE_PORT_REF(A)
#endif
#if defined (PORTB)
AVRLIB_SPECIALIZE_PORT_REF(B)
#endif
#if defined (PORTC)
AVRLIB_SPECIALIZE_PORT_REF(C)
#endif
#if defined (PORTD)
AVRLIB_SPECIALIZE_PORT_REF(D)
#endif
#if defined (PORTE)
AVRLIB_SPECIALIZE_PORT_REF(E)
#endif
#if defined (PORTF)
AVRLIB_SPECIALIZE_PORT_REF(F)
#endif
#if defined (PORTG)
AVRLIB_SPECIALIZE_PORT_REF(G)
#endif


//      
inline uint8_t bit_mask(__bits bit) { return _BV((uint8_t)(bit)); }
inline uint8_t bit_mask(uint8_t bit) { return bit_mask((__bits)bit); }

//      
inline void setbit(volatile uint8_t& ref, __bits bit) { ref |= bit_mask(bit); }
inline void setbit(volatile uint8_t& ref, uint8_t bit) { setbit(ref, (__bits)bit); }
inline void clrbit(volatile uint8_t& ref, __bits bit) { ref &= ~bit_mask(bit); }
inline void clrbit(volatile uint8_t& ref, uint8_t bit) { clrbit(ref, (__bits)bit); }

/**
 *   /
 **/

template <__ports port, __bits bit>
class __pin
{
public:
	static inline void set(void)
	{
		__ports_ref<port, __reg_PORT>()|=bit_mask(bit);
	}
	static inline void clr(void)
	{
		__ports_ref<port, __reg_PORT>()&=~bit_mask(bit);
	}

	static  void toggle(void)
	{
		is_set() ? clr():set();
	}

	static inline bool is_set(void)
	{
		return (__ports_ref<port, __reg_PIN>() & bit_mask(bit)) != 0;
	}
	static inline bool is_clear(void)
	{
		return (__ports_ref<port, __reg_PIN>() & bit_mask(bit)) == 0;
	}
	static inline void as_input(void)
	{
		__ports_ref<port, __reg_DDR>() &=~bit_mask(bit);
	}
	static inline void as_output(void)
	{
		__ports_ref<port, __reg_DDR>() |=bit_mask(bit);
	}
	static inline void pull_up_on(void)
	{
		__ports_ref<port, __reg_PORT>()|=bit_mask(bit);
	}
	static inline void pull_up_off(void)
	{
		__ports_ref<port, __reg_PORT>()&=~bit_mask(bit);
	}

};

template <__ports port, __bits bit>
class __in
{
public:

	static inline bool is_set(void)
	{
		return (__ports_ref<port, __reg_PIN>() & bit_mask(bit)) != 0;
	}
	static inline bool is_clear(void)
	{
		return (__ports_ref<port, __reg_PIN>() & bit_mask(bit)) == 0;
	}
	static inline void pull_up_on(void)
	{
		__ports_ref<port, __reg_PORT>()|=bit_mask(bit);
	}
	static inline void pull_up_off(void)
	{
		__ports_ref<port, __reg_PORT>()&=~bit_mask(bit);
	}

};

template <__ports port, __bits bit>
class __out
{
public:
	static inline void set(void)
	{
		__ports_ref<port, __reg_PORT>()|=bit_mask(bit);
	}
	static inline void clr(void)
	{
		__ports_ref<port, __reg_PORT>()&=~bit_mask(bit);
	}

	static inline void toggle(void)
	{
		if(is_set())
		{
			clr();
		}
		else
		{
			set();
		}
	}

	static inline bool is_set(void)
	{
		return (__ports_ref<port, __reg_PIN>() & bit_mask(bit)) != 0;
	}
	static inline bool is_clear(void)
	{
		return (__ports_ref<port, __reg_PIN>() & bit_mask(bit)) == 0;
	}

};

template <__ports port>
class __port
{
public:
	static inline void set_DDR(uint8_t ddr_mask)
	{
		__ports_ref<port, __reg_DDR>()&=ddr_mask;
	}
	static inline void set_pull_up(uint8_t pull_up_mask)
	{
		__ports_ref<port, __reg_PORT>()&=pull_up_mask;
	}
	static inline void set_port(uint8_t port_)
	{
		__ports_ref<port, __reg_PORT>()&=port_;
	}
	static inline void init(uint8_t ddr_mask,uint8_t pull_up_mask)
	{
		__ports_ref<port, __reg_PORT>()&=pull_up_mask;
		__ports_ref<port, __reg_DDR>()&=ddr_mask;
	}

	static inline uint8_t pins(void)
	{
		return __ports_ref<port, __reg_PIN>();
	}

	static __pin<port,__bit_0> pin_0;
	static __pin<port,__bit_1> pin_1;
	static __pin<port,__bit_2> pin_2;
	static __pin<port,__bit_3> pin_3;
	static __pin<port,__bit_4> pin_4;
	static __pin<port,__bit_5> pin_5;
	static __pin<port,__bit_6> pin_6;
	static __pin<port,__bit_7> pin_7;

};

class atomic
{
public:
	atomic() : status_reg(SREG) {cli();}
	~atomic(){SREG=status_reg;}

private:
	uint8_t status_reg;
};

class non_atomic
{
public:
	non_atomic() : status_reg(SREG) {sei();}
	~non_atomic(){SREG=status_reg;}

private:
	uint8_t status_reg;
};

template <typename safe_var_type>
class safe_var
{
public:
	void set(safe_var_type data)
	{
		atomic cs;
		var=data;
	}
	safe_var_type get(void)
	{
		atomic cs;
		return var;
	}
	inline void unsafe_set(safe_var_type data)
	{
		var=data;
	}
	inline 	safe_var_type unsafe_get(void)
	{
		return var;
	}
private:
	volatile safe_var_type var;
};

}

#endif /* AVRIO_HPP_ */
